/***************************************************************************
 *
 * Copyright (c) 2000, 2001, 2002 BalaBit IT Ltd, Budapest, Hungary
 * All rights reserved.
 *
 * $Id: tpsocket.c,v 1.9.6.1 2002/08/26 07:56:34 sasa Exp $
 *
 * Author  : Bazsi
 * Auditor :
 * Last audited version:
 * Notes:
 *
 ***************************************************************************/

#include <zorp/zorp.h>
#include <zorp/tpsocket.h>
#include <zorp/log.h>
#include <zorp/cap.h>
#include <zorp/nfiptproxy-kernel.h>

#include <stdlib.h>


#if ENABLE_NETFILTER_TPROXY

gchar *auto_bind_ip = "1.2.3.4";

static gint
z_tp_autobind(gint fd)
{
  struct sockaddr_in sa;
  
  memset(&sa, 0, sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_port = htons(0);
  z_inet_aton(auto_bind_ip, &sa.sin_addr);
  /* */
  if (bind(fd, (struct sockaddr *) &sa, sizeof(sa)) == -1)
    {
      z_log(NULL, CORE_ERROR, 3, "Error autobinding transparent socket; fd='%d', ip='%s', error='%m'", fd, auto_bind_ip);
      errno = EADDRNOTAVAIL;
      return -1;
    }
  return 0;
}


static gint
z_do_tp_bind(gint fd, struct sockaddr *sa, socklen_t salen)
{
  struct in_tproxy itp;
  socklen_t itplen = sizeof(itp);
  
  if (sa->sa_family != AF_INET)
    return z_do_ll_bind(fd, sa, salen);
  
  if (bind(fd, sa, salen) != -1)
    {
      /* we could bind successfully, it was a local address */

      /* we don't want NAT, only that -m tproxy matches our session */
      itp.itp_fport = 0;
      itp.itp_faddr.s_addr = 0;
      if (setsockopt(fd, SOL_IP, IP_TPROXY_ASSIGN, &itp, itplen) == -1)
	{
          z_log(NULL, CORE_ERROR, 3, "Error in setsockopt(SOL_IP, IP_TPROXY_ASSIGN), netfilter-tproxy support required; fd='%d', error='%m'", fd);
	  return -1;
	}
      return 0;
    }
  else if (errno != EADDRNOTAVAIL)
    return -1;
    
  /* we couldn't bind because the address was not available, try transparent proxy tricks */
  if (z_tp_autobind(fd) == -1)
    {
      /* autobind was not successful either */
      return -1;
    }
  itp.itp_fport = ((struct sockaddr_in *) sa)->sin_port;
  itp.itp_faddr = ((struct sockaddr_in *) sa)->sin_addr;
  if (setsockopt(fd, SOL_IP, IP_TPROXY_ASSIGN, &itp, itplen) == -1)
    {
      z_log(NULL, CORE_ERROR, 3, "Error in setsockopt(SOL_IP, IP_TPROXY_ASSIGN), netfilter-tproxy support required; fd='%d', error='%m'", fd);
      return -1;
    }
  return 0;
}

static gint
z_do_tp_listen(gint fd, gint backlog, gboolean accept_one)
{
  unsigned int flags;
  socklen_t flagslen = sizeof(flags);
  
  z_enter();
  if (z_do_ll_listen(fd, backlog, accept_one) == -1)
    {
      z_leave();
      return -1;
    }
  if (getsockopt(fd, SOL_IP, IP_TPROXY_FLAGS, &flags, &flagslen) == -1)
    {
      z_leave();
      return 0;
    }
  if (flags & ITP_MARK)
    return 0;
    
  flags = ITP_LISTEN | (accept_one ? ITP_ONCE : 0);
  if (setsockopt(fd, SOL_IP, IP_TPROXY_FLAGS, &flags, sizeof(flags)) == -1)
    {
      z_log(NULL, CORE_ERROR, 3, "Error in setsockopt(SOL_IP, IP_TPROXY_FLAGS), netfilter-tproxy support required; fd='%d', error='%m'", fd);
      z_leave();
      return -1;
    }
  z_leave();
  return 0;
}

static gint
z_do_tp_connect(gint fd, struct sockaddr *sa, socklen_t salen)
{
  unsigned int flags;
  struct in_tproxy itp;
  socklen_t itplen = sizeof(itp);
  
  if (getsockopt(fd, SOL_IP, IP_TPROXY_QUERY, &itp, &itplen) == -1)
    {
      return z_do_ll_connect(fd, sa, salen);
    }
  flags = ITP_CONNECT | ITP_ONCE;
  if (setsockopt(fd, SOL_IP, IP_TPROXY_FLAGS, &flags, sizeof(flags)) == -1)
    {
      z_log(NULL, CORE_ERROR, 3, "Error in setsockopt(SOL_IP, IP_TPROXY_FLAGS), netfilter-tproxy support required; fd='%d', error='%m'", fd);
      return -1;
    }
  return z_do_ll_connect(fd, sa, salen);
}
  
static gint
z_do_tp_getsockname(gint fd, struct sockaddr *sa, socklen_t *salen)
{
  struct in_tproxy itp;
  socklen_t itplen = sizeof(itp);
  struct sockaddr_in *sin = (struct sockaddr_in *) sa;
      
  if (getsockopt(fd, SOL_IP, IP_TPROXY_QUERY, &itp, &itplen) == -1)
    {
      return z_do_ll_getsockname(fd, sa, salen);
    }
  if (*salen < sizeof(*sin))
    {
      errno = EFAULT;
      return -1;
    }
  sin->sin_family = AF_INET;
  sin->sin_addr = itp.itp_faddr;
  sin->sin_port = itp.itp_fport;
  *salen = sizeof(*sin);  
  return 0;
}

ZSocketFuncs z_tp_socket_funcs = 
{
  z_do_tp_bind,
  z_do_ll_accept,
  z_do_tp_connect,
  z_do_tp_listen,
  z_do_tp_getsockname,
  z_do_ll_getpeername
};

void
z_tp_socket_init(void)
{
  if (system("/sbin/modprobe iptable_tproxy") == 0)
    socket_funcs = &z_tp_socket_funcs;
}

#else

void
z_tp_socket_init(void)
{
}

#endif
